"""
**yodel** (*the Swiss Army knife of your sound*) is an easy-to-use python
package for digital audio signal processing, analysis and synthesis.
"""

__author__ = 'Romain Clement'
__copyright__ = 'Copyright 2014, Romain Clement'
__credits__ = []
__license__ = 'MIT'
__version__ = '0.1.1'
__maintainer__ = 'Romain Clement'
__email__ = 'contact@romainclement.com'
__status__ = 'Development'



import math
import array


class DFT:
    """
    The Discrete Fourier Transform allows to convert a time-domain signal
    into a frequency-domain spectrum.

    *Warning:*
        It should not be used in practice for computational reasons,
        and should only be used for testing purposes. Instead, prefer
        using the :py:class:`.FFT`.

    *Reference:*
        "Digital Signal Processing, a practical guide for engineers and
        scientists", Steven W. Smith
    """

    def __init__(self, size):
        """
        Initialize the Discrete Fourier Transform.

        :param size: length of the DFT (should only be a power of 2)
        """
        self.size = size
        self._generate_lookup_tables()

    def _generate_lookup_tables(self):
        """
        Generate internal lookup tables for trigonometric functions (sin, cos)
        """
        table_size = self.size * self.size
        self.cos_table = array.array('f', [0] * table_size)
        self.sin_table = array.array('f', [0] * table_size)
        two_pi = 2.0 * math.pi
        for i in range(0, table_size):
            self.cos_table[i] = math.cos(two_pi * i / self.size)
            self.sin_table[i] = math.sin(two_pi * i / self.size)

    def forward(self, real_signal, real_spec, imag_spec):
        """
        Compute the complex spectrum of a given real time-domain signal

        :param real_signal: real time-domain input signal
        :param real_spec: real-part of the output complex spectrum
        :param imag_spec: imaginary-part of the output complex spectrum
        """
        for i in range(0, self.size):
            real_spec[i] = 0
            imag_spec[i] = 0

        for k in range(0, self.size):
            for i in range(0, self.size):
                real_spec[k] += real_signal[i] * self.cos_table[k * i]
                imag_spec[k] -= real_signal[i] * self.sin_table[k * i]

    def inverse(self, real_spec, imag_spec, real_signal):
        """
        Compute the real time-domain signal of a given complex spectrum

        :param real_spec: real-part of the complex spectrum
        :param imag_spec: imaginary-part of the complex spectrum
        :param real_signal: real time-domain output signal
        """
        for i in range(0, self.size):
            real_signal[i] = 0

        for k in range(0, self.size):
            real_spec[k] = real_spec[k] / self.size
            imag_spec[k] = - imag_spec[k] / self.size

        for k in range(0, self.size):
            for i in range(0, self.size):
                real_signal[i] += real_spec[k] * self.cos_table[k * i]
                real_signal[i] += imag_spec[k] * self.sin_table[k * i]


class FFT:
    """
    The Fast Fourier Transform is a faster algorithm for performing the
    :py:class:`.DFT`. It allows converting a time-domain signal into a
    frequency-domain spectrum.

    *Reference:*
        "Digital Signal Processing, a practical guide for engineers and
        scientists", Steven W. Smith
    """

    def __init__(self, size):
        """
        Initialize the Fast Fourier Transform.

        :param size: length of the FFT (should only be a power of 2)
        """
        self.size = size
        self._generate_lookup_tables()

    def _generate_lookup_tables(self):
        """
        Generate internal lookup tables for trigonometric functions (sin, cos)
        """
        table_size = self.size
        self.cos_table = array.array('f', [0] * table_size)
        self.sin_table = array.array('f', [0] * table_size)
        for i in range(1, table_size):
            self.cos_table[i] = math.cos(math.pi / i)
            self.sin_table[i] = math.sin(math.pi / i)

    def forward(self, real_signal, real_spec, imag_spec):
        """
        Compute the complex spectrum of a given real time-domain signal

        :param real_signal: real time-domain input signal
        :param real_spec: real-part of the output complex spectrum
        :param imag_spec: imaginary-part of the output complex spectrum
        """
        spec_len = int(self.size / 2)

        for i in range(0, spec_len):
            real_spec[i] = real_signal[2 * i]
            imag_spec[i] = real_signal[2 * i + 1]

        n = spec_len

        nm1 = n - 1
        nd2 = int(n / 2)
        m = round(math.log(n) / math.log(2))
        j = nd2

        for i in range(1, nm1):
            if i < j:
                tr = real_spec[j]
                ti = imag_spec[j]
                real_spec[j] = real_spec[i]
                imag_spec[j] = imag_spec[i]
                real_spec[i] = tr
                imag_spec[i] = ti
            k = nd2
            while k <= j:
                j -= k
                k /= 2
            j = int(j + k)

        for l in range(1, int(m + 1)):
            le = round(math.pow(2, l))
            le2 = int(le / 2)
            ur = 1
            ui = 0
            sr = self.cos_table[le2]
            si = - self.sin_table[le2]

            for j in range(1, int(le2 + 1)):
                jm1 = j - 1
                for i in range(int(jm1), int(nm1 + 1), int(le)):
                    ip = i + le2
                    tr = real_spec[int(ip)] * ur - imag_spec[int(ip)] * ui
                    ti = real_spec[int(ip)] * ui + imag_spec[int(ip)] * ur
                    real_spec[int(ip)] = real_spec[i] - tr
                    imag_spec[int(ip)] = imag_spec[i] - ti
                    real_spec[i] += tr
                    imag_spec[i] += ti
                tr = ur
                ur = tr * sr - ui * si
                ui = tr * si + ui * sr

        n = n * 2

        nm1 = n - 1
        nd2 = int(n / 2)
        n4 = int(n / 4 - 1)

        for i in range(1, n4 + 1):
            im = nd2 - i
            ip2 = i + nd2
            ipm = im + nd2
            real_spec[ip2] = (imag_spec[i] + imag_spec[im]) / 2
            real_spec[ipm] = real_spec[ip2]
            imag_spec[ip2] = -(real_spec[i] - real_spec[im]) / 2
            imag_spec[ipm] = -imag_spec[ip2]
            real_spec[i] = (real_spec[i] + real_spec[im]) / 2
            real_spec[im] = real_spec[i]
            imag_spec[i] = (imag_spec[i] - imag_spec[im]) / 2
            imag_spec[im] = -imag_spec[i]

        real_spec[int(n * 3 / 4)] = imag_spec[int(n / 4)]
        real_spec[nd2] = imag_spec[0]
        imag_spec[int(n * 3 / 4)] = 0
        imag_spec[nd2] = 0
        imag_spec[int(n / 4)] = 0
        imag_spec[0] = 0

        l = round(math.log(n) / math.log(2))
        le = round(math.pow(2, l))
        le2 = int(le / 2)
        ur = 1
        ui = 0
        sr = self.cos_table[le2]
        si = - self.sin_table[le2]

        for j in range(1, int(le2 + 1)):
            jm1 = j - 1
            for i in range(jm1, nm1 + 1, int(le)):
                ip = int(i + le2)
                tr = real_spec[ip] * ur - imag_spec[ip] * ui
                ti = real_spec[ip] * ui + imag_spec[ip] * ur
                real_spec[ip] = real_spec[i] - tr
                imag_spec[ip] = imag_spec[i] - ti
                real_spec[i] += tr
                imag_spec[i] += ti
            tr = ur
            ur = tr * sr - ui * si
            ui = tr * si + ui * sr

    def inverse(self, real_spec, imag_spec, real_signal):
        """
        Compute the real time-domain signal of a given complex spectrum

        :param real_spec: real-part of the complex spectrum
        :param imag_spec: imaginary-part of the complex spectrum
        :param real_signal: real time-domain output signal
        """
        n = self.size
        nspec = int(n / 2 + 1)

        for k in range(nspec, n):
            real_spec[k] = real_spec[n - k]
            imag_spec[k] = - imag_spec[n - k]

        for k in range(0, n):
            real_spec[k] = real_spec[k] + imag_spec[k]

        tmp_real = [0] * n
        tmp_imag = [0] * n

        self.forward(real_spec, tmp_real, tmp_imag)

        for i in range(0, n):
            real_signal[i] = (tmp_real[i] + tmp_imag[i]) / n


class AnalysisWindow:
    """
    An analysis window function allows to reduce unwanted frequencies
    when performing spectrum analysis.
    """

    def __init__(self, size):
        """
        Initialize the analysis window. By default, a flat window is applied.
        Use one of the provided methods to make it Hanning or Hamming.

        :param size: length of the analysis window
        """
        self._resize(size)

    def hanning(self, size):
        """
        Make a Hanning analysis window.

        :param size: length of the analysis window
        """
        self._resize(size)
        for i in range(0, self.size):
            self.signal[i] = (0.5 -
                              (0.5 * math.cos(2.0 * math.pi * i /
                                              (self.size - 1))))

    def hamming(self, size):
        """
        Make a Hamming analysis window.

        :param size: length of the analysis window
        """
        self._resize(size)
        for i in range(0, self.size):
            self.signal[i] = (0.54 -
                              (0.46 * math.cos(2.0 * math.pi * i /
                                               (self.size - 1))))

    def blackman(self, size):
        """
        Make a Blackman analysis window.

        :param size: length of the analysis window
        """
        self._resize(size)
        for i in range(0, self.size):
            self.signal[i] = (0.42659 -
                              (0.49656 * math.cos(2.0 * math.pi * i /
                                                  (self.size - 1))) +
                              (0.076849 * math.cos(4.0 * math.pi * i /
                                                   (self.size - 1))))

    def process(self, input_signal, output_signal):
        """
        Perform windowing on an input signal.

        :param input_signal: input signal to be windowed
        :param output_signal: resulting windowed signal
        """
        for i in range(0, self.size):
            output_signal[i] = input_signal[i] * self.signal[i]

    def _resize(self, size):
        """
        Resize the analysis window.

        :param size: new length of the analysis window
        """
        self.size = size
        self.signal = array.array('f', [1] * self.size)



import math


def modulus(re, im):
    """
    Compute the modulus of a complex number.

    :param re: real part of the complex number
    :param im: imaginary part of the complex number
    :rtype: modulus of complex number
    """
    return math.sqrt(re * re + im * im)


def phase(re, im):
    """
    Compute the phase of a complex number.

    :param re: real part of the complex number
    :param im: imaginary part of the complex number
    :rtype: phase of complex number
    """
    return math.atan2(im, re)



import math


def lin2db(lin):
    """
    Convert a linear value to the decibel (dB) scale.

    :param lin: linear value
    :rtype: decibel value
    """
    if lin > 1e-5:
        return 20.0 * math.log10(lin)
    else:
        return -100.0


def db2lin(db):
    """
    Convert a decibel (dB) value to the linear scale.

    :param db: decibel value
    :rtype: linear value
    """
    return math.pow(10, db / 20.0)



from setuptools import setup
from setuptools.command.test import test as TestCommand
import sys
import os
import re
import codecs

here = os.path.abspath(os.path.dirname(__file__))

def read(*parts):
    # intentionally *not* adding an encoding option to open
    return codecs.open(os.path.join(here, *parts), 'r').read()

def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string")

class Tox(TestCommand):

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import tox
        errcode = tox.cmdline(self.test_args)
        sys.exit(errcode)

setup(
    name='yodel',
    version=find_version('yodel', '__init__.py'),
    description='The Swiss Army knife of your sound',
    long_description=read('README.md'),
    author='Romain Clement',
    author_email='contact@romainclement.com',
    url='https://github.com/rclement/yodel',
    packages=['yodel'],
    tests_require=['tox'],
    cmdclass={'test': Tox},
    platforms='any',
    license='MIT'
)



import math
import array


class Biquad:
    """
    A biquad filter is a 2-poles/2-zeros filter allowing to perform
    various kind of filtering. Signal attenuation is at a rate of 12 dB
    per octave.

    *Reference:*
        "Cookbook formulae for audio EQ biquad filter coefficients",
        Robert Bristow-Johnson
        (http://www.musicdsp.org/files/Audio-EQ-Cookbook.txt)
    """

    def __init__(self):
        """
        Create an inactive biquad filter with a flat frequency response.
        To make the filter active, use one of the provided methods.
        """
        self.reset()

    def reset(self):
        """
        Make the filter inactive with a flat frequency response.
        """
        self._a_coeffs = array.array('f', [0] * 3)
        self._b_coeffs = array.array('f', [0] * 3)
        self._x1 = 0.0
        self._x2 = 0.0
        self._y1 = 0.0
        self._y2 = 0.0

    def low_pass(self, samplerate, cutoff, resonance):
        """
        Make a low-pass filter.

        :param samplerate: sample-rate in Hz
        :param cutoff: cut-off frequency in Hz
        :param resonance: resonance or Q-factor
        """
        self._compute_constants(samplerate, cutoff, resonance)
        self._a_coeffs[0] = (1.0 + self._alpha)
        self._a_coeffs[1] = (-2.0 * self._cos_w0) / self._a_coeffs[0]
        self._a_coeffs[2] = (1.0 - self._alpha) / self._a_coeffs[0]
        self._b_coeffs[0] = ((1.0 - self._cos_w0) / 2.0) / self._a_coeffs[0]
        self._b_coeffs[1] = (1.0 - self._cos_w0) / self._a_coeffs[0]
        self._b_coeffs[2] = ((1.0 - self._cos_w0) / 2.0) / self._a_coeffs[0]

    def high_pass(self, samplerate, cutoff, resonance):
        """
        Make a high-pass filter.

        :param samplerate: sample-rate in Hz
        :param cutoff: cut-off frequency in Hz
        :param resonance: resonance or Q-factor
        """
        self._compute_constants(samplerate, cutoff, resonance)
        self._a_coeffs[0] = (1.0 + self._alpha)
        self._a_coeffs[1] = (-2.0 * self._cos_w0) / self._a_coeffs[0]
        self._a_coeffs[2] = (1.0 - self._alpha) / self._a_coeffs[0]
        self._b_coeffs[0] = ((1.0 + self._cos_w0) / 2.0) / self._a_coeffs[0]
        self._b_coeffs[1] = -(1.0 + self._cos_w0) / self._a_coeffs[0]
        self._b_coeffs[2] = ((1.0 + self._cos_w0) / 2.0) / self._a_coeffs[0]

    def band_pass(self, samplerate, center, resonance):
        """
        Make a band-pass filter.

        :param samplerate: sample-rate in Hz
        :param center: center frequency in Hz
        :param resonance: resonance or Q-factor
        """
        self._compute_constants(samplerate, center, resonance)
        self._a_coeffs[0] = (1.0 + self._alpha)
        self._a_coeffs[1] = (-2.0 * self._cos_w0) / self._a_coeffs[0]
        self._a_coeffs[2] = (1.0 - self._alpha) / self._a_coeffs[0]
        self._b_coeffs[0] = (self._alpha) / self._a_coeffs[0]
        self._b_coeffs[1] = 0
        self._b_coeffs[2] = (- self._alpha) / self._a_coeffs[0]

    def all_pass(self, samplerate, center, resonance):
        """
        Make an all-pass filter.

        :param samplerate: sample-rate in Hz
        :param center: center frequency in Hz
        :param resonance: resonance or Q-factor
        """
        self._compute_constants(samplerate, center, resonance)
        self._a_coeffs[0] = (1.0 + self._alpha)
        self._a_coeffs[1] = (-2.0 * self._cos_w0) / self._a_coeffs[0]
        self._a_coeffs[2] = (1.0 - self._alpha) / self._a_coeffs[0]
        self._b_coeffs[0] = (1.0 - self._alpha) / self._a_coeffs[0]
        self._b_coeffs[1] = (-2.0 * self._cos_w0) / self._a_coeffs[0]
        self._b_coeffs[2] = (1.0 + self._alpha) / self._a_coeffs[0]

    def notch(self, samplerate, center, resonance):
        """
        Make a notch filter.

        :param samplerate: sample-rate in Hz
        :param center: center frequency in Hz
        :param resonance: resonance or Q-factor
        """
        self._compute_constants(samplerate, center, resonance)
        self._a_coeffs[0] = (1.0 + self._alpha)
        self._a_coeffs[1] = (-2.0 * self._cos_w0) / self._a_coeffs[0]
        self._a_coeffs[2] = (1.0 - self._alpha) / self._a_coeffs[0]
        self._b_coeffs[0] = (1.0) / self._a_coeffs[0]
        self._b_coeffs[1] = (-2.0 * self._cos_w0) / self._a_coeffs[0]
        self._b_coeffs[2] = (1.0) / self._a_coeffs[0]

    def peak(self, samplerate, center, resonance, dbgain):
        """
        Make a peak filter.

        :param samplerate: sample-rate in Hz
        :param center: center frequency in Hz
        :param resonance: resonance or Q-factor
        :param dbgain: gain in dB
        """
        self._compute_constants(samplerate, center, resonance, dbgain)
        self._a_coeffs[0] = (1.0 + self._alpha / self._a)
        self._a_coeffs[1] = (-2.0 * self._cos_w0) / self._a_coeffs[0]
        self._a_coeffs[2] = (1.0 - self._alpha / self._a) / self._a_coeffs[0]
        self._b_coeffs[0] = (1.0 + self._alpha * self._a) / self._a_coeffs[0]
        self._b_coeffs[1] = (-2.0 * self._cos_w0) / self._a_coeffs[0]
        self._b_coeffs[2] = (1.0 - self._alpha * self._a) / self._a_coeffs[0]

    def low_shelf(self, samplerate, cutoff, resonance, dbgain):
        """
        Make a low-shelf filter.

        :param samplerate: sample-rate in Hz
        :param cutoff: cut-off frequency in Hz
        :param resonance: resonance or Q-factor
        :param dbgain: gain in dB
        """
        self._compute_constants(samplerate, cutoff, resonance, dbgain)
        self._a_coeffs[0] = ((self._a + 1) +
                             (self._a - 1) * self._cos_w0 + self._sqrtAlpha)
        self._a_coeffs[1] = ((-2.0 * ((self._a - 1) +
                             (self._a + 1) * self._cos_w0)) /
                             self._a_coeffs[0])
        self._a_coeffs[2] = (((self._a + 1) +
                              (self._a - 1) * self._cos_w0 - self._sqrtAlpha) /
                             self._a_coeffs[0])
        self._b_coeffs[0] = ((self._a *
                              ((self._a + 1) -
                               (self._a - 1) * self._cos_w0 +
                               self._sqrtAlpha)) /
                             self._a_coeffs[0])
        self._b_coeffs[1] = ((2.0 * self._a *
                              ((self._a - 1) - (self._a + 1) * self._cos_w0)) /
                             self._a_coeffs[0])
        self._b_coeffs[2] = ((self._a *
                              ((self._a + 1) -
                               (self._a - 1) * self._cos_w0 -
                               self._sqrtAlpha)) /
                             self._a_coeffs[0])

    def high_shelf(self, samplerate, cutoff, resonance, dbgain):
        """
        Make a high-shelf filter.

        :param samplerate: sample-rate in Hz
        :param cutoff: cut-off frequency in Hz
        :param resonance: resonance or Q-factor
        :param dbgain: gain in dB
        """
        self._compute_constants(samplerate, cutoff, resonance, dbgain)
        self._a_coeffs[0] = ((self._a + 1) -
                             (self._a - 1) * self._cos_w0 + self._sqrtAlpha)
        self._a_coeffs[1] = ((2.0 *
                              ((self._a - 1) - (self._a + 1) * self._cos_w0)) /
                             self._a_coeffs[0])
        self._a_coeffs[2] = (((self._a + 1) -
                              (self._a - 1) * self._cos_w0 -
                              self._sqrtAlpha) /
                             self._a_coeffs[0])
        self._b_coeffs[0] = ((self._a *
                              ((self._a + 1) +
                               (self._a - 1) * self._cos_w0 +
                               self._sqrtAlpha)) /
                             self._a_coeffs[0])
        self._b_coeffs[1] = ((-2.0 * self._a *
                              ((self._a - 1) + (self._a + 1) * self._cos_w0)) /
                             self._a_coeffs[0])
        self._b_coeffs[2] = ((self._a *
                              ((self._a + 1) +
                               (self._a - 1) * self._cos_w0 -
                               self._sqrtAlpha)) /
                             self._a_coeffs[0])

    def custom(self, a0, a1, a2, b0, b1, b2):
        """
        Make a custom filter.

        :param a0: a[0] coefficient
        :param a1: a[1] coefficient
        :param a2: a[2] coefficient
        :param b0: b[0] coefficient
        :param b1: b[1] coefficient
        :param b2: b[2] coefficient
        """
        self._a_coeffs[0] = a0
        self._a_coeffs[1] = a1 / a0
        self._a_coeffs[2] = a2 / a0
        self._b_coeffs[0] = b0 / a0
        self._b_coeffs[1] = b1 / a0
        self._b_coeffs[2] = b2 / a0

    def process_sample(self, x):
        """
        Filter a single sample and return the filtered sample.

        :param x: input sample
        :rtype: filtered sample
        """
        curr = x
        y = (self._b_coeffs[0] * x +
             self._b_coeffs[1] * self._x1 +
             self._b_coeffs[2] * self._x2 -
             self._a_coeffs[1] * self._y1 -
             self._a_coeffs[2] * self._y2)
        self._x2 = self._x1
        self._x1 = curr
        self._y2 = self._y1
        self._y1 = y
        return y

    def process(self, x, y):
        """
        Filter an input signal. Can be used for in-place filtering.

        :param x: input buffer
        :param y: output buffer
        """
        num_samples = len(x)
        for n in range(0, num_samples):
            y[n] = self.process_sample(x[n])

    def _compute_constants(self, fs, fc, q, dbgain=0):
        """
        Pre-compute internal mathematical constants
        """
        self._fc = fc
        self._q = q
        self._dbgain = dbgain
        self._fs = fs
        self._a = math.pow(10, (self._dbgain / 40.0))
        self._w0 = 2.0 * math.pi * self._fc / self._fs
        self._cos_w0 = math.cos(self._w0)
        self._sin_w0 = math.sin(self._w0)
        self._alpha = self._sin_w0 / (2.0 * self._q)
        self._sqrtAlpha = 2.0 * math.sqrt(self._a) * self._alpha
